##################################################################################
##
## FUNCTIONS FOR CLASS CLUSTER-ALGORITHM-CLARA
##
##################################################################################

clusterclara <- function(dataset,
                         clusters,
                         distance = "euclidean",
                         repeats = 1,
                         fixed = TRUE)
{
  clusterlist <- list()
            
  cat("Clustering the dataset.\n")
            
  for (j in 1:repeats)
              {
                c <- clara(dataset,
                           clusters,
                           rngR = !fixed,
                           metric = distance)
                clusterlist <- c(clusterlist, split(seq(1:dim(dataset)[1]),
                                                    as.factor(c$clustering)))
              }
            
  d <- mean(silhouette(c))
  
  if (abs(d - clusters) > 0.2 * clusters)
    {
      cat("CLARA suggests",
          round(d, 0),
          "clusters. You requested",
          clusters,
          "clusters. You might want to adapt your settings.\n")
    }
  
  cat("Finished...\n")

  clusterlist
}

##################################################################################
##
## CLASS CLUSTER-ALGORITHM-CLARA
##
##################################################################################

setClass("clusterAlgorithmClara",
         representation(
                        distance = "character",
                        clusters = "numeric",
                        fixed    = "logical",
                        repeats  = "numeric"
                        ),
         prototype(distance = "euclidian",
                   clusters = 50,
                   fixed = FALSE,
                   config = list(
                     distance =
                     function (object)
                     {
                       selectOption(c("euclidean", "manhattan"),
                                    paste("Please select the distance measure to be\n",
                                          "used. See the CLARA documentation for a detailed\n",
                                          "explanation (?clara).", sep ="")
                                    )
                     },
                     clusters =
                     function (object)
                     {
                       selectNumber(paste("Please select the number of clusters\n",
                                          "for the CLARA clustering.", sep =""), 2, 10000)
                     },
                     fixed =
                     function (object)
                     {
                       selectOption(c("fixed", "not fixed"),
                                    paste("Please select whether clara should always result\n",
                                          "in the same clustering (fixed)\nor whether you\n",
                                          "want to take real random numbers into account\n",
                                          "so that CLARA works non-deterministically (not\n",
                                          "fixed).For further help consult the clara\n",
                                          "documentation (?clara).", sep = "")
                                    )=="fixed"
                     },
                     repeats =
                     function (object)
                     {
                       if (!object@fixed)
                         {
                           return(selectNumber("Please select how often the CLARA clustering should be repeated.", 2, 1000))
                         }
                       return(1)
                     }
                     ),
                   setup = list(
                     distance =
                     function (X, object)
                     {
                       if (length(X %in% c("euclidean", "manhattan")) == 0)
                         stop("The distance matrix for clara has to be either \"euclidian\" or \"manhattan\"!")
                     },
                     clusters =
                     function (X, object)
                     {
                       if (as.integer(X) < 2)
                         stop(paste("The \"clusters\" value is missing or to low!\n",
                                    "Please specify an integer larger than 2!", sep = "")
                              )
                     },
                     fixed =
                     function (X, object)
                     {
                       if (!is.logical(X))
                         stop(paste("The random number generation for Clara has\n",
                                    "to be set to FALSE (\"result not fixed\")\n",
                                    "or TRUE (\"fixed result\")!", sep = "")
                              )
                     },
                     repeats =
                     function (X, object)
                     {
                       if (!object@fixed)
                         {
                           if (as.integer(X) < 1)
                             stop(paste("The \"repeats\" value is missing! You activated\n",
                                        "the random number generation so you should also\n",
                                        "set how often the clustering should be repeated!",
                                        sep = "")
                                  )
                         }
                     }
                     ),
                   execute = function(object, parent)
                   {
                     if (checkData(object@clusterset, "clustered", "execute", TRUE)) return
                     
                     ## Requirement for Clara
                     if (!require(cluster)) stop("Please install the package cluster for the clara function!")
                     
                     if (is.null(parent))
                       stop("The object needs to have a defined parent before it can be used!")

                     object@clusterset <- clusterclara(exprs(parent@data@dataset),
                                                       object@clusters,
                                                       object@distance,
                                                       object@repeats,
                                                       object@fixed
                                                       )
                     object
                   }
                   ),
         contains = "clusterAlgorithm"
         )

setMethod("print", "clusterAlgorithmClara", function(x, ...)
          {
            printData("Algorithm",
                      "Clara",
                      c(paste("Distance measure:", x@distance),
                        paste("Number of clusters:", x@clusters),
                        ifelse(length(x@clusterset)!=0, "The dataset has been clustered.", "The object holds no result yet.")))
          }
          )

