##################################################################################
##
## FUNCTIONS FOR CLASS CLUSTER-ALGORITHM-HCLUST
##
##################################################################################

HCLUSTMETHOD   <- c("complete", "ward", "single", "average",
                    "mcquitty", "median", "centroid")
HCLUSTDISTANCE <- c("euclidean", "maximum", "manhattan",
                    "canberra","binary","minkowski")

clusterhclust <- function(dataset,
                          method = "complete",
                          distance = "euclidean")
{

  ## Converts the hclust tree into a list of list as
  ## needed by goCluster
  getGroups <- function(X)
    {
      if (   !is.null(attr(X[[1]],"leaf"))
          && !is.null(attr(X[[2]],"leaf")))
        {
          return(list(c(X[[1]], X[[2]])))
        }
      else if (!is.null(attr(X[[1]],"leaf")))
        {
          Y <- getGroups(X[[2]])
          return(list(Y,c(Y[[length(Y)]], X[[1]])))
        }
      else if (!is.null(attr(X[[2]],"leaf")))
        {
          Y <- getGroups(X[[1]])
          return(list(Y,c(Y[[length(Y)]], X[[2]])))
        }
      else
        {
          Y1 <- getGroups(X[[1]])
          Y2 <- getGroups(X[[2]])
          return(list(Y1, Y2, c(Y1[[length(Y1)]], Y2[[length(Y2)]])))
        }
    }
            
  
  cat("Clustering the dataset.\n")
  
  d <- dist(dataset, method = distance)
  
  h <- hclust(d, method = method)

  list(hclust   = h,
       listtree =  getGroups(as.dendrogram(h)))
}

##################################################################################
##
## CLASS CLUSTER-ALGORITHM-HCLUST
##
##################################################################################

setClass("clusterAlgorithmHclust",
         representation(
                        tree = "list",
                        method = "character",
                        distance = "character"
                        ),
         prototype(method = "complete",
                   distance="euclidian",
                   config = list(
                     method =
                     function (object)
                     {
                       selectOption(HCLUSTMETHOD,
                                    paste("Please select the agglomeration method\n",
                                          "that should be used. For details see the\n",
                                          "documentation to the \"hclust\" function (?hclust).",
                                          sep = "")
                                    )
                     },
                     distance =
                     function (object)
                     {
                       selectOption(HCLUSTDISTANCE,
                                    paste("Please select the distance measure to be\n",
                                          "used. For details see the documentation to\n",
                                          "the \"hclust\" function (?hclust).", sep = "")
                                    )
                     }
                   ),
                   setup = list(
                     method =
                     function (X, object)
                     {
                       if (!any(X == HCLUSTMETHOD))
                         stop(paste("You did not specify the agglomeration method\n",
                                    "to be used. This can be one of ",
                                    paste(HCLUSTMETHOD, collapse="\", \"", sep =""),
                                    ". For details see the \"hclust\" function.",
                                    sep = "")
                              )
                     },
                     distance =
                     function (X, object)
                     {
                       if (!any(X == HCLUSTDISTANCE))
                         stop(paste("You did not specify the distance measure to be\n",
                                    "used. This can be one of ",
                                    paste("\"", HCLUSTDISTANCE, "\"", collapse="\", \"", sep =""),
                                    ". For details see the \"hclust\" function.", sep ="")
                              )
                     }
                     ),
                   execute = function(object, parent)
                   {
                     if (checkData(object@clusterset, "clustered", "execute", TRUE)) return
                     
                     if (is.null(parent))
                       stop("The object needs to have a defined parent before it can be used!")

                     result <- clusterhclust(exprs(parent@data@dataset),
                                             object@method,
                                             object@distance)
                     
                     ## FIXME: This is a problem on extracting the object later
                     class(result$hclust) <- "list"

                     ## safe the tree for later reference
                     object@tree <- result$hclust

                     object@clusterset <- result$listtree
                     
                     
                     object
                   }                     
                   ),
         contains = "clusterAlgorithm"
         )

setMethod("print", "clusterAlgorithmHclust", function(x, ...)
          {
            printData("Algorithm",
                      "Hclust",
                      c(paste("Distance measure:", x@distance),
                        paste("Clustering method:", x@method),
                        ifelse(length(x@clusterset)!=0, "The dataset has been clustered.", "The object holds no result yet.")))
          }
          )
