##################################################################################
##
## FUNCTIONS FOR CLASS CLUSTER-ANNOTATION-CHROMOSOME-ABP
##
##################################################################################

prepareCHRannotation <- function(uniqueID,
                                 annBuilderPkg,
                                 chromosomes = NULL,
                                 trueID = NULL)
  {
    ## Taken from GOstat (Bioconductor)
    getDataEnv <- function(name, lib) {
      get(paste(lib, name, sep=""), mode="environment")
    }
    
    ## this is a good time to load the required package
    if (!require(annBuilderPkg, character.only = TRUE))
      stop(paste("Please install the AnnBuilder meta package \"", annBuilderPkg,
                 "\" that you specified as annotation ressource!", sep ==""))
            
    cat("Building chromosome annotation...\n")
                     
    ## Get all unique ids available
    gidlist <- ls(env = getDataEnv("CHR", annBuilderPkg))
    
    ## reduce the list of gene ids by the ones actually
    ## present in the dataset
    gidlist <- gidlist[!is.na(match(gidlist,
                                    uniqueID))]
    
    ## Create a table with chromosome information and gene id 
    chrtable <- cbind(as.character(mget(gidlist, env = getDataEnv("CHR", annBuilderPkg))), gidlist)
    ## Reject duplicated entries (if a true id is given, this should be taken
    ## as basis to determine duplicated entries)
    if (!is.null(trueID) && trueID != "")
      {
        chrtable[, 2] <- as.character(unlist(mget(chrtable[, 2],
                                                  env = getDataEnv(trueID, annBuilderPkg),
                                                  ifnotfound = NA)))
        
        gidlist <- as.character(unlist(mget(gidlist,
                                            env = getDataEnv(trueID, annBuilderPkg),
                                            ifnotfound = NA)))
      
        uniqueID <- as.character(unlist(mget(uniqueID,
                                             env = getDataEnv(trueID, annBuilderPkg),
                                             ifnotfound = NA)))
      }
    
    chrtable <- chrtable[!duplicated(paste(chrtable[, 1], chrtable[, 2])), ]
    
    cat("Done!\n")
  
    list(annotations = list(chr = as.matrix(chrtable)),
         uniqueid    = uniqueID)
  }

checkannotationpackage <- function(X, Y)
{
  if (!require(X, character.only = TRUE))
    {
      cat(paste("The package \"", X, "\" could not be loaded! Did you install it?", sep = ""))
      return(FALSE)
    }
      envr <- paste(X, Y, sep ="")
      for (i in envr)
        {
          if (!exists(i)) 
            {
              cat(paste("The environment \"", i, "\" does not exist! It should be part of the meta package \"", X, "\".", sep = ""))
              return(FALSE)
            }
          if (!is.environment(get(i))) 
            {
              cat(paste("The object \"", i, "\" is no environment! Is \"", X, "\" a Bioconductor meta package?", sep = ""))
              return(FALSE)
            }
        }
    
  return(TRUE)
}

checkchrannotation <- function(X)
{
  checkannotationpackage(X, c("CHR"))
}


##################################################################################
##
## CLASS CLUSTER-ANNOTATION-GOab
##
##################################################################################

setClass("clusterAnnotationChromosome-abp",
         representation(
                        ## The environment that contains
                        ## GO annotation information
                        meta       = "character",
                        ## The linking annotation
                        trueid     = "character",
                        ## A (possibly transformed) copy of the data unique id
                        uniqueid   = "character",
                        ## The chromosomes you are interested in
                        chromosomes= "character"
                        ),
         prototype(type = "Chromosome",
                   config = list(
                     meta =
                     function (object)
                     {
                       selectCore(paste("Please specify the name of the Bioconductor\n",
                                        "meta package that applies to your dataset.\n",
                                        "This has to be an AnnBuilder package that\n",
                                        "holds all information to your array.\n",
                                        "If you are using Affymetrix arrays there is\n",
                                        "a high chance you will find a prepackaged\n",
                                        "dataset at the Bioconductor website and for\n",
                                        "spotted array you will have to create your\n",
                                        "own AnnBuilder package.", sep =""),
                                  check = checkchrannotation
                                  )
                       
                     },
                     trueid =
                     function (object)
                     {
                       selectCore(paste("If the unique id you specified for the\n",
                                        "dataset is not the true linker between\n",
                                        "genes and annotation, please specify the\n",
                                        "linking annotation here. This is the\n",
                                        "case for packages like hgu95av2 where the\n",
                                        "affymetrix probe ids are the ids that\n",
                                        "link all annotations together, but the locus\n",
                                        "link ids are the actual ids that have been\n",
                                        "used to establish the package in the first\n",
                                        "place. The name you specify here needs to\n",
                                        "yield the name of an environment together with\n",
                                        "the package you specified the step before.\n",
                                        "If you don't specify anthing this option is disabled."),
                                  check = function(X) {
                                    if (X != "") {
                                      return(exists(paste(object@meta, X, sep ="")))
                                    } else {
                                      return(TRUE)
                                    }
                                  }
                                  )
                     },
                     chromosomes =
                     function (object)
                     {
                       chr <- names(table(unlist(mget(ls(env = get(paste(object@meta,
                                                           "CHR", sep =""))),
                                                      env = get(paste(object@meta,
                                                        "CHR", sep =""))))))
            
                       selectOptions(chr,
                                     paste("Please choose the chromosomes you\n",
                                           "want to include in the analysis.\n",
                                           "If you use the commandline configuration\n",
                                           "you can select several chromosomes by \n",
                                           "seperating them with a comma\n",
                                           "or by specifying ALL to include all.\n",
                                        sep = ""
                                        ),
                                     "ALL"
                                     )
                     }
                     ),
                   setup = list(
                     meta =
                     function(X, object)
                     {
                       if (!checkchrannotation(X))
                         stop("Invalid annotation data package!")
                     },
                     trueid = 
                     function(X, object)
                     {
                       if (!is.character(X) || !length(X) > 0)
                         stop("The true linker ID should be a string!")
                     },
                     chromosomes =
                     function(X, object)
                     {
                       chr <- names(table(unlist(mget(ls(env = get(paste(object@meta,
                                                           "CHR", sep =""))),
                                                      env = get(paste(object@meta,
                                                        "CHR", sep =""))))))
                       
                       a <- intersect(X, chr)
                       
                       if(length(a) == 0)
                         stop(paste("Invalid chromosome specification (", X, ")!", sep = ""))

                     }
                     ),
                   execute = function(object, parent)
                   {
                     
                     if (length(object@annoset)>0)
                       {
                         stop("The object has already been annotated! Ignoring \"execute\" call!")
                       }
                     
                     if (is.null(parent))
                       stop("The object needs to have a defined parent before it can be used!")
                     
                     if (!extends(class(parent),"goCluster"))
                       stop(paste("The specified parent needs to be an object of class \"goCluster\" or a derived class, but is instead an object of class \"", class(parent),"\"!", sep = ""))
                     
                     if (is.null(object@meta))
                       {
                         stop("The class has not been setup correctly! The meta package information is missing.")
                       }

                     result <- prepareCHRannotation(parent@data@uniqueid,
                                                   object@meta,
                                                   object@chromosomes,
                                                   object@trueid)
                     
                     object@uniqueid <- result[["uniqueid"]]
                     
                     object@annoset <- result[["annotations"]]
                                          
                     object

                   }
                   ),
         contains = "clusterAnnotation"
         )

setMethod("print", "clusterAnnotationChromosome-abp", function(x, ...)
          {
            printData("Annotation",
                      "Chromosome",
                      c(paste("Chromosome localization retrieved from meta package:", x@meta),
                        paste("Chromosomes of interest:", paste(x@chromosomes, collapse = ", "))))
          }
          )

