##################################################################################
##
## FUNCTIONS FOR CLASS CLUSTER-ANNOTATION-GO-ABP
##
##################################################################################

EVIDENCECODES <- c("IC", "IDA", "IEA", "IEP",
                   "IGI", "IMP", "IPI", "ISS",
                   "NAS", "ND", "TAS", "NR")

ONTOLOGIES    <- c("MF","BP","CC")

prepareGOannotation <- function(uniqueID,
                                annBuilderPkg,
                                ontologies = ONTOLOGIES,
                                evidence = EVIDENCECODES,
                                trueID = NULL)
{

  ## this is a good time to load the required packages
  if (!require(GO))
    stop("Please install the GO package to be able to parse the GO annotation!")
  if (!require(annBuilderPkg, character.only = TRUE))
    stop(paste("Please install the AnnBuilder meta package \"", annBuilderPkg,
               "\" that you specified as annotation ressource!", sep ==""))
            
  ## Taken from GOstat (Bioconductor)
  getDataEnv <- function(name, lib) {
    get(paste(lib, name, sep=""))
  }
            
  cat("Building GO annotation...\n")

  ## Get all unique ids available
  denv <- getDataEnv("GO", annBuilderPkg)
  gidlist <- ls(env = denv)

  ## reduce the list of gene ids by the ones actually
  ## present in the dataset
  gidlist <- gidlist[!is.na(match(gidlist,
                                  uniqueID))]

  ## Get the reverse mapping of GO terms to gene ids
  go2all <- unlist(as.list(getDataEnv("GO2ALLPROBES", annBuilderPkg)))

  ## Create a table with GO id, gene id and GO evidence
  go2alltable <- cbind(substr(names(go2all), 1, 10),
                       go2all,
                       substr(names(go2all), 12, nchar(names(go2all)))
                       )
  ## All annotations without evidence code are given "ND"
  go2alltable[is.na(match(go2alltable[, 3], EVIDENCECODES)), 3] <- "ND" 
  ## Remove annotations based on evidence that is not acceptable
  go2alltable <- go2alltable[!is.na(match(go2alltable[, 3], evidence)), ]
  ## remap all id lists to the true id
  if (!is.null(trueID) && trueID != "")
    {
      ## FIXME: This will create a problem if one uniqueID matches several trueIDs
      go2alltable[, 2] <- as.character(unlist(mget(go2alltable[, 2],
                                                   env = getDataEnv(trueID, annBuilderPkg),
                                                   ifnotfound = NA)))
      
      gidlist <- as.character(unlist(mget(gidlist,
                                          env = getDataEnv(trueID, annBuilderPkg),
                                          ifnotfound = NA)))
      
      uniqueID <- as.character(unlist(mget(uniqueID,
                                           env = getDataEnv(trueID, annBuilderPkg),
                                           ifnotfound = NA)))
    }

  ## Reject duplicated entries 
  go2alltable <- go2alltable[!duplicated(paste(go2alltable[, 1],
                                               go2alltable[, 2])), ]
  
  ## remove missing values
  matchgenego <- match(go2alltable[, 2], gidlist)
  go2alltable <- go2alltable[!is.na(matchgenego),]
  ## select the ontologies
  annotations <- list()
  for (i in ontologies)
    {
      matchontogo <- match(go2alltable[, 1],
                           ls(env = getDataEnv(
                                paste(i,"CHILDREN", sep =""),
                                "GO")
                              ))
      ## save the terms in the annoset (we dont need the evidence codes anymore)
      annotations[[i]] <- go2alltable[!is.na(matchontogo), 1:2]
      ## check if we actually have any GO terms left in this ontology
      if (nrow(annotations[[i]]) == 0)
        warning(paste("Ontology",i,"has no GO terms left."))
    }
  
  cat("Done!\n")
  
  list(annotations = annotations,
       uniqueid    = uniqueID)
}

checkgoannotation <- function(X)
{
  checkannotationpackage(X, c("GO", "GO2ALLPROBES"))
}

##################################################################################
##
## CLASS CLUSTER-ANNOTATION-GO-ABP
##
##################################################################################

setClass("clusterAnnotationGO-abp",
         representation(
                        ## The environment that contains
                        ## GO annotation information
                        meta       = "character",
                        ## The linking annotation
                        trueid     = "character",
                        ## The accepted evidence codes
                        evidence   = "character",
                        ## The ontologies to analyze
                        ontologies = "character",
                        ## A (possibly transformed) copy of the data unique id
                        uniqueid   = "character"
                        ),
         prototype(config = list(
                     meta =
                     function (object)
                     {
                       selectCore(paste("Please specify the name of the Bioconductor\n",
                                        "meta package that applies to your dataset.\n",
                                        "This has to be an AnnBuilder package that\n",
                                        "holds all information to your array.\n",
                                        "If you are using Affymetrix arrays there is\n",
                                        "a high chance you will find a prepackaged\n",
                                        "dataset at the Bioconductor website and for\n",
                                        "spotted array you will have to create your\n",
                                        "own AnnBuilder package.", sep =""),
                                  check = checkgoannotation
                                  )
                       
                     },
                     trueid =
                     function (object)
                     {
                       selectCore(paste("If the unique id you specified for the",
                                        "\ndataset is not the true linker between",
                                        "\ngenes and annotation, please specify the",
                                        "\nlinking annotation here. This is the",
                                        "\ncase for packages like hgu95av2 where the",
                                        "\naffymetrix probe ids are the ids that",
                                        "\nlink all annotations together, but the locus",
                                        "\nlink ids are the actual ids that have been",
                                        "\nused to establish the package in the first",
                                        "\nplace. The name you specify here needs to",
                                        "\nyield the name of an environment together with",
                                        "\nthe package you specified the step before.",
                                        "\nIf you don't specify anthing this option is disabled."),
                                  check = function(X) {
                                    if (X != "") {
                                      return(exists(paste(object@meta, X, sep ="")))
                                    } else {
                                      return(TRUE)
                                    }
                                  }
                                  )
                     },
                     ontologies =
                     function (object)
                     {
                       selectOptions(ONTOLOGIES,
                                     paste("Please choose the type of ontology to use.",
                                           "\nYou can use the following abbreviations:",
                                           "\nMolecular function (MF),",
                                           "\nBiological Process (BP),",
                                           "\nCellular component (CC),",
                                           "\nall three (GO).",
                                           "\n(You can select several ontologies",
                                           "\nby separating them with a comma in",
                                           "\ncase you use the commandline configuration)"
                                           ),
                                     "GO"
                                     )
                     },
                     evidence =
                     function (object)
                     {
                       selectOptions(EVIDENCECODES,
                                     paste("Please choose the type of GO annotations",
                                           "\nthat you are willing to accept:",
                                           "\nIC:  inferred by curator,",
                                           "\nIDA: inferred from direct assay,",
                                           "\nIEA: inferred from electronic annotation,",
                                           "\nIEP: inferred from expression pattern,",
                                           "\nIGI: inferred from genetic interaction,",
                                           "\nIMP: inferred from mutant phenotype,",
                                           "\nIPI: inferred from physical interaction,",
                                           "\nISS: inferred from sequence or structural similarity,",
                                           "\nNAS: non-traceable author statement,",
                                           "\nND:  no biological data available,",
                                           "\nTAS: traceable author statement,",
                                           "\nNR:  not recorded,",
                                           "\nALL: accept all types given above.",
                                           "\nIf you need more information on the",
                                           "\ndifferent types, please refer to this page:",
                                           "\nhttp://www.geneontology.org/GO.evidence.html",
                                           "\n(You can select several evidence codes",
                                           "\nby separating them with a comma in",
                                           "\ncase you use the commandline configuration)"
                                           ),
                                     "ALL"
                                     )
                     }
                     ),
                   setup = list(
                     meta =
                     function (X, object)
                     {
                       if (!checkgoannotation(X))
                         stop("Invalid annotation data package!")
                     },
                     trueid = 
                     function(X, object)
                     {
                       if (!is.character(X) || !length(X) > 0)
                         stop("The true linker ID should be a string!")
                     },
                     ontologies = 
                     function(X, object)
                     {
                       a <- ONTOLOGIES[match(X, ONTOLOGIES)]
                       if(length(a) == 0) stop(paste("Invalid ontology specification (", paste(X, collapse = ", "),")",sep=""))
                     },
                     evidence = 
                     function(X, object)
                     {
                       a <- EVIDENCECODES[match(X, EVIDENCECODES)]
                       if(length(a) == 0) stop(paste("Invalid evidence specification (",paste(X, collapse = ", "),")",sep=""))
                     }
                     ),
                   execute = function(object, parent)
                   {
                     if (is.null(parent))
                       stop("The object needs to have a defined parent before it can be used!")
                     
                     if (!extends(class(parent),"goCluster"))
                       stop(paste("The specified parent needs to be an object of class \"goCluster\" or a derived class, but is instead an object of class \"", class(a),"\"!", sep = ""))
                     
                     if (is.null(object@meta))
                       stop("The class has not been setup correctly! The meta package information is missing.")

                     object@ontologies <- intersect(object@ontologies, ONTOLOGIES)
                     if (length(object@ontologies) == 0)
                       stop("No valid ontologies defined.")
                       
                     
                     object@evidence <- intersect(object@evidence, EVIDENCECODES)
                     if (length(object@evidence) == 0) 
                       stop("No valid evidence codes defined.")
                     
                     result <- prepareGOannotation(parent@data@uniqueid,
                                                   object@meta,
                                                   object@ontologies,
                                                   object@evidence,
                                                   object@trueid)
                     
                     object@uniqueid <- result[["uniqueid"]]
                     
                     object@annoset <- result[["annotations"]]
                                          
                     object
                   },
                   type = "GO"
                   ),
         contains = "clusterAnnotation"
         )


setMethod("print", "clusterAnnotationGO-abp", function(x, ...)
          {
            printData("Annotation",
                      "Gene Ontology (GO)",
                      c(paste("GO information retrieved from meta package", x@meta),
                        paste("GO ontologies selected:", paste(x@ontologies, collapse = ", ")),
                        paste("Accepted type of evidence:", paste(x@evidence, collapse = ", "))))
          }
          )

