##################################################################################
##
## FUNCTIONS FOR CLASS CLUSTER-SIGNIF-FDR
##
##################################################################################

randomizeTree <- function(X, elements)
{
  if (is.list(X))
    {
      return(lapply(X, randomizeTree, elements))
    }
  else
    {
      return(round(runif(length(X),min=1,max=elements),0))
    }
}

mergeAnno <- function(X, i)
{
  if (length(X)==0) return(NA)
  if (is.list(X[[1]]))
    {
      return(lapply(X, mergeAnno, i))
    }
  else
    {
      return(X[[i]])
    }
}

selectAnnoStats <- function(X, threshold)
{
  if (length(X)==0) return(NA)
  if (is.list(X[[1]]))
    {
      a <- lapply(X, selectAnnoStats, threshold)
      return(a)
    }
  else
    {
      a <- list()
      for (i in 1:length(X))
        {
          b <- X[[i]][ X[[i]] < threshold[i] ]
          if (is.null(b))
            {
              a[[i]] <- NA
            }
          else
            {
              a[[i]] <- b
              if (length(a[[i]]) == 0) a[[i]] <- NA
            }
        }
      return(a)
    }
}

##################################################################################
##
## CLASS CLUSTER-SIGNIF-FDR
##
##################################################################################

setClass("clusterSignifFDR",
         representation(
                        repeats     = "numeric",
                        threshold   = "numeric",
                        randomstat  = "list",
                        fdr         = "list",
                        pthresholds = "numeric"
                        ),
         prototype(threshold = 0.05,
                   repeats = 100,
                   config = list(
                     repeats =
                     function (object)
                     {
                       selectNumber(paste("Please select how often randomization",
                                          "\nshould be performed."), 0, 10000)
                     },
                     threshold =
                     function (object)
                     {
                       selectDouble(paste("Please select the false discovery rate",
                                          "\nfor the selection of GO terms."), 0, 1)
                     }
                     ),
                   setup = list(
                     threshold =
                     function (X, object)
                     {
                       if (as.double(X) == 0)
                         stop(paste("The \"threshold\" value is missing! Please",
                                    "\nspecify a threshold for your significance",
                                    "\nanalysis!"))
                     },
                     repeats =
                     function (X, object)
                     {
                       if (as.integer(X) == 0)
                         stop(paste("The \"repeats\" value is missing! Please",
                                    "\nspecify how often the data should be",
                                    "\nrandomized!"))
                     }
                     ),
                   execute =
                     function (object, parent)
                     {
                       ##FIXME: Export as a single function?
                       copyParent <- parent
                       
                       randomGroups <- list()

                       cat("Randomizing groups.\n")
                       
                       for (i in 1:object@repeats)
                         {
                           cat(round(i/object@repeats*100),"%..")
                           randomGroups[[i]] <- randomizeTree(parent@algo@clusterset,
                                                              length(parent@data@anno@uniqueid))
                         }
                       
                       copyParent@algo@clusterset <- randomGroups

                       cat("\nAnalyzing random data.\n")

                       object@randomstat <- execute(object@stat, copyParent)@statset
                       
                       randomStats <- list()

                       for (i in 1:length(parent@data@anno@annoset))
                         {
                           cat(round(i/length(parent@data@anno@annoset)*100),"%..")
                           randomStats[[i]] <- unlist(mergeAnno(object@randomstat, i))
                         }
                       
                       cat("\nAnalyzing original data.\n")
                       
                       object@stat    <- execute(object@stat, parent)
                       object@statset <- object@stat@statset

                       realStats <- list()

                       for (i in 1:length(parent@data@anno@annoset))
                         {
                           realStats[[i]] <- unlist(mergeAnno(object@statset, i))
                         }

                       FDR <- list()
                       threshold <- numeric()

                       for (i in 1:length(realStats))
                         {
                           realStats[[i]] <- unlist(mergeAnno(object@statset, i))

                           a  <- 0
                           b  <- 0
                           
                           xx <- cbind(c(randomStats[[i]],realStats[[i]]),
                                       c(rep(TRUE, length(randomStats[[i]])),
                                         rep(FALSE, length(realStats[[i]]))))

                           xx <- xx[order(xx[,1]),]
                           
                           for (j in 1:dim(xx)[1])
                             {
                               if (xx[j, 2])
                                 a <- a + 1
                               else
                                 b <- b + 1

                               xx[j, 2] <- a/b/object@repeats
                             }

                           xy <- seq(1,dim(xx)[1])[xx[,2] < object@threshold]

                           if (length(xy) == 0)
                             {
                               threshold[i] <- 0
                             }
                           else
                             {
                               threshold[i] <- xx[max(xy) , 1]
                             }

                           FDR[[i]] <- xx
                           
                         }

                       cat("Selecting significant annotations.\n")

                       object@fdr <- FDR
                       object@pthresholds <- threshold
                       object@selection <- selectAnnoStats(object@statset, threshold)
                       
                       object
                       
                     }

           ),
         contains = "clusterSignif"
         )


setMethod("print", "clusterSignifFDR", function(x, ...)
          {
            if (length(x@selection)!=0)
              {
                nrselected <- unlist(x@selection)
                nrselected <- nrselected[!is.na(nrselected)]
                report <- c("The clustered data has been analyzed.",
                            paste("A total of", length(nrselected), "annotation terms have been selected")
                            )
              }
            else
              {
                report <- "The object holds no result yet."
              }
            printData("Significance",
                      "False Discovery Rate",
                      report
                      )
            print(x@stat)
          }
          )
