##################################################################################
##
## FUNCTIONS FOR CLASS CLUSTER-VISUAL-HCLUST
##
##################################################################################

minimizeSelection <- function(X, annotations) {

  ## This selects the entry with the minimal value
  minselect <- function(Y) (Y[Y == min(Y)])[1]

  branchlength <- length(X)

  ## if the given branch X has as many items as the number of given annotations
  if (branchlength == annotations) 
    {
      ## and no element is a list
      if (all(unlist(lapply(X, function(Y) !is.list(Y)))))
        {
          ## then we are at the bottom of the tree, will select the minimal
          ## annotation term in each entry and reduce the list to a vector
          return(unlist(lapply(X, minselect)))
        }
    }

  lowest <- list()
  for (i in 1:length(X))
    {
      Y <- X [[i]]
      if (is.list(Y))
        {
          ## follow a branch in the tree
          lowest[[i]] <- minimizeSelection(Y, annotations)
        }
      else
        {
          ## select minimal p-value for a final element
          lowest[[i]] <- unlist(lapply(Y, minselect))
        }
    }
  return(lowest)
}

mergeResult <- function(result, mergedResult, annotations) {

  for (i in 1:annotations)
    {
      A <- result[[i]]
      B <- mergedResult[[i]]
      ## intersect the gene id
      overlap <- intersect(A[, 1], B[, 1])
      ## no overlap?
      if (length(overlap) == 0)
        {
          ## return joined result
          mergedResult[[i]] <- rbind(A, B)
        }
      else
        {
          ## for each identical gene id
          for (j in overlap)
            {
              new <- A[A[, 1] == j, ]
              old <- B[B[, 1] == j, ]
              ## both p-values are present and the new one is lower than the old one
              if (!is.na(new[2]) && (is.na(old[2]) || new[2] < old[2]))
                {
                  ## replace the line in the old array by the line in the new one
                  B[B[, 1] == j, ] <- A[A[, 1] == j, ]
                } 
            }
          ## merge all non-overlapping gene-ids to the old array
          mergedResult[[i]] <- rbind(B, A[-match(overlap, A[, 1]), ])
        }
    }
  return(mergedResult)
}

prepareAxis <- function(minSelection, Clusters, annotations)
{

  mergedResult <- NULL

  ## process each item in the list
  for (i in 1:length(minSelection))
    {
      ## Is the current item branching?
      if (is.list(minSelection[[i]])) 
        {
          ## follow the branches
          result <- prepareAxis(minSelection[[i]], Clusters[[i]], annotations)
        }
      else
        {
          result <- list()
          ## consider the p-value of each annotation
          for (j in 1:annotations)
            {
              ## read the p-value
              pvalue <- minSelection[[i]][j]
              ## remember the annotation term
              aterm <- names(minSelection[[i]])[j]
              ## fix for NA p-values
              if (is.null(aterm)) aterm <- NA
              ## generate a matrix that contains the gene ids from the
              ## clustering result, the corresponding annotation term
              ## and the p-value
              result[[j]] <- cbind(Clusters[[i]],
                                   rep(pvalue, length(Clusters[[i]])),
                                   rep(aterm, length(Clusters[[i]])))
            }
        }
      ## if there is already a result present we need to merge it with the
      ## new one in order to select the minimal, non-overlapping regions
      if (is.null(mergedResult))
        {
          mergedResult <- result
        }
      else
        {
          mergedResult <- mergeResult(result, mergedResult, annotations)
        }
    }

  return(mergedResult)
}

##################################################################################
##
## CLASS CLUSTER-VISUAL-HCLUST
##
##################################################################################

setClass("clusterVisualHclust",
         representation(
                        prepared = "list"
                        ),
         contains = "clusterVisual",
         prototype = list(
           execute =
           function (object, parent)
           {
             ## get the number of annotations available
             annotations <- length(parent@data@anno@annoset)
             ## check the tree of selected items for the minimal p-values
             minSelection <- minimizeSelection(parent@sign@selection, annotations)
             ## merge the tree selections to an axis that contains only the minimal, non-overlapping regions
             object@prepared <- prepareAxis(minSelection, parent@algo@clusterset, annotations)

             object
           }
           )
         )

setMethod("print", "clusterVisualHclust", function(x, ...)
          {
            if (length(x@prepared)!=0)
              {
                report <- c("The clustered data has been prepared for visualization.",
                            paste("A total of", length(x@prepared), "annotation types can be visualized.","\n")
                            )
              }
            else
              {
                report <- "The object holds no result yet."
              }
            printData("Visualization",
                      "Hclust",
                      report
                      )
          }
          )

setMethod("display", "clusterVisualHclust", function(object, parent, selection, lorder, ...)
          {

            if (missing(selection))
              stop(paste("You need to select one of the", length(parent@data@anno@annoset),"annotation(s) that are available (e.g. specify selection=\"1\" in the function call to select the first annotation).\n"))

              if (class(parent@algo)!="clusterAlgorithmHclust")
              stop("Sorry this visualization is meant to be used with a goCluster object that uses hierarchical clustering.")
            
            if (is.null(object@prepared[[selection]]))
              stop(paste("You selected an annotation that does not exist. Please specify a value between 1 and ", length(parent@data@anno@annoset),".\n"))
            
            genecolors <- getGeneColors()

            tree <- parent@algo@tree
            class(tree) <- "hclust"
            dendro <- as.dendrogram(tree)
            
            if (!missing(lorder))
              dendro <- reorder( dendro  , lorder )
            
            gorder <- order.dendrogram(dendro)

            x <- exprs(parent@data@dataset)[gorder, ]
            
            layout(matrix(1:2, nrow = 1, ncol = 2, byrow = TRUE), widths = c(2, 2), height = c(1), respect = FALSE )
          
            par(cex.axis = 1.1, cex.main = 2)

            par(mar = c(5, 0.5, 0, 0) + 0.1)
          
            image(1:ncol(x), 
                  1:nrow(x), 
                  t(x), 
                  axes = FALSE, 
                  xlim = c(0.5,ncol(x) + 0.5), 
                  ylim = c(0.5, nrow(x) + 0.5), 
                  zlim = c(min(x, na.rm = TRUE) - 0.5, max(x, na.rm = TRUE) + 0.5),
                  xlab = "Samples", 
                  ylab = "", 
                  col  = rgb(
                    genecolors[1,],
                    genecolors[2,],
                    genecolors[3,],
                    maxColorValue = 255),
                  main = "", 
                  cex.main = 1.5)
            
            axis(1, 1:ncol(x), 
                 las = 2, 
                 line = -0.5, 
                 tick = 0, 
                 labels = colnames(x)
                 )
            
            box()
            
            axis <- object@prepared[[selection]]

            plot.new()
            
            par(usr=c(0,3,0,nrow(x)))

            orderedAxis <- (axis[order(as.integer(axis[,1])),])[gorder,]
            
            plotAxis(orderedAxis, ...)

          }
          )

plotAxis <- function(axis,
                     hide,
                     mark,
                     line,
                     spacing,
                     linespace,
                     minSlot,
                     minSize,
                     adjust)
{

  if (missing(line)) line <- 0.15
  if (missing(mark)) mark <- ""
  if (missing(hide)) hide <- ""
  if (missing(spacing)) spacing <- 15
  if (missing(linespace)) linespace <- 0.13
  if (missing(minSlot)) minSlot <- 20
  if (missing(minSize)) minSize <- 8 
  if (missing(adjust)) adjust <- 20

  ## Number of genes
  genes <- dim(axis)[1]

  if(any(is.na(axis[,3])))
    {
      axis[is.na(axis[,3]),2] <- 1
      axis[is.na(axis[,3]),3] <- "NA-invalid"
      hide <- c(hide,"NA-invalid")
    }
  
  ## determine breakpoints (switch of annotation term)
  breakpoints <- seq(1, genes - 1)[axis[1:(genes - 1), 3] != axis[2:genes, 3]]

  c1   <- 0 
  linespace <- 0.13
  tp <- 1
  ts <- 1.3
  markcol  <- "#FEE090"
  
  for (c2 in c(breakpoints, genes))
    {
      ## 
      if (   (c2 - c1) > minSize
          && !is.na(axis[c2, 3])
          && !any(axis[c2, 3] == hide))
        {
          middlepoint <- mean(c(c1, c2))
          if (middlepoint < minSlot)
            textpoint <- minSlot
          else
            textpoint <- ifelse(middlepoint - adjust < minSlot,  minSlot, middlepoint - adjust)
          if (any(axis[c2, 3] == mark)) 
            {
              rect(tp + 0.05, textpoint - 1.5 * spacing, 
                   3, textpoint + 0.5 * spacing, 
                   col = markcol, border = FALSE)
              sc1 <- markcol
              sc2 <- markcol
            }
          else
            {
              sc1 <- "black"
              sc2 <- "lightgrey"
            }
          lines(c(line, tp),   c(middlepoint, textpoint), lwd = 1, col = sc2)
          lines(c(line, line), c(c1, c2), lwd = 2, col = sc1)
          text(tp + 0.1, textpoint, axis[c2, 3], 0, cex = ts)
      
          minSlot <- textpoint + 2 * spacing
          line <- ifelse(line >= (4 * linespace), linespace, line + linespace)
        }
      c1 <- c2 + 1
    }
}
