require(Biobase)
##################################################################################
##
## GOCLUSTER FUNCTION DEFINITIONS
##
##################################################################################

## Retrieves classes that extend the specified class in the given
## environment (default is global env)

getChildrenClasses <- function(X, where = .GlobalEnv)
  {
    Y <- character()
    
    for(i in getClasses(where = where, inherit = TRUE))
      {
        if(extends(i, X) & X != i)
          {
            Y <- c(Y, i)
          }
      }
    Y
  }


selectCore  <- function(message,
                        convert = function(X) X,
                        check = function(X) TRUE,
                        options = NULL,
                        optflag = TRUE,
                        error = function(X) return("Invalid choice!"),
                        ...
                        )
  {
    if (options()$goCluster.useTk)
      {
        if(require(tcltk))
          {
            selectCoreTk(message, convert, check, options, error, ...)
          }
        else
          {
            warn("You need to install the tcltk package to get the graphical user interface! Reverting to text interface.")
            selectCorePlain(message, convert, check, options, optflag, error, ...)
          }
      }
    else
      {
        selectCorePlain(message, convert, check, options, optflag, error, ...)
      }
  }

## A generic function
selectCorePlain <- function(message,
                            convert = function(X) X,
                            check = function(X) TRUE,
                            options = NULL,
                            optflag = TRUE,
                            error = function(X) return("Invalid choice!"),
                            ...
                            )
  {
    ## Disply the message to the user
    cat(message,"\n")

    ## Any optional arguments?
    optArgs <- list(...)

    ## Function to reduce the optional arguments to the ones
    ## actually needed for the specified function
    addArgs <- function(fnc, args) {
      X <- match(names(formals(fnc)), names(args))
      X[!is.na(X)]
    }

    ## Set argument lists 
    cnvArgs <- optArgs[addArgs(convert, optArgs)]
    chkArgs <- optArgs[addArgs(check,   optArgs)]
    errArgs <- optArgs[addArgs(error,   optArgs)]
    
    while (TRUE) {
      ## Display options if there are any
      if (!is.null(options))
        {
          if (!optflag) 
            {
              cat("[ ", paste(options, collapse = ","), " ]\n")
            }
          else
            {
              for (i in 1:length(options))
                cat(i,")", options[i], "\n")
            }
        }

      ## Read user input
      X <- readline("\n# ")
      
      if (!optflag && !is.null(options)) X <- match(unlist(strsplit(X, ",")), options)
      
      ## convert to the required format
      choice <- do.call("convert",c(list(X), cnvArgs))
      cat("\n")

      ## did the conversion fail?
      if (any(is.na(choice))) {
        cat(do.call("error",c(list(X = choice), errArgs)), "\n")
        next
      }
      ## Check the user choice. If the function returns TRUE,
      ## the user input is ok and can be returned
      if (do.call("check",c(list(X = choice), chkArgs))) break
      
      cat(do.call("error",c(list(X = choice), errArgs)), "\n")
    } 
    choice
  }

selectCoreTk <- function(message,
                         convert = function(X) X,
                         check = function(X) TRUE,
                         options = NULL,
                         optflag = TRUE,
                         error = function(X) return("Invalid choice!"),
                         ...
                         )
  {

    ## Taken from fileBrowser()
    LABELFONT1 <- "Helvetica 12 bold"
    LABELFONT2 <- "Helvetica 11"

    doAnswer <- function()
      {
        uChoice <<- NULL
      }
    end <- function()
      {
        doAnswer()
        tkgrab.release(ubox)
        tkdestroy(ubox)
      }

    
    ## Any optional arguments?
    optArgs <- list(...)

    ## Function to reduce the optional arguments to the ones
    ## actually needed for the specified function
    addArgs <- function(fnc, args) {
      X <- match(names(formals(fnc)), names(args))
      X[!is.na(X)]
    }

    ## Set argument lists 
    cnvArgs <- optArgs[addArgs(convert, optArgs)]
    chkArgs <- optArgs[addArgs(check,   optArgs)]
    errArgs <- optArgs[addArgs(error,   optArgs)]
    
    while (TRUE) {
      ## Start the dialog
      ubox <- tktoplevel()
      
      ## Set the dialog title
      tktitle(ubox) <- "goCluster configuration"
      
      ## Disply the message to the user
      uMessage <- tklabel(ubox, text = message, font = LABELFONT1)
      tkpack(uMessage)
      
      ## Display options if there are any
      if (!is.null(options))
        {
          selectedOptions <- character()
          for (i in 1:length(options))
            {
              if (optflag)
                {
                  if (i==1) sval <- "active" else sval <- "normal"
                  uSelection <- tkradiobutton(ubox, text = options[i], font = LABELFONT2, variable = "selection", value = i, state = sval)
                }
              else
                {
                  uSelection <- tkcheckbutton(ubox, text = options[i], font = LABELFONT2, offvalue = -1, onvalue = i, variable = paste("checkoption", i, sep = ""))
                }
              tkpack(uSelection)
            }
          if (optflag)
            {
              doAnswer <- function()
                {
                  uChoice <<- tclvalue("selection")
                }
            }
          else
            {
              doAnswer <- function()
                {
                  uChoice <<- numeric()
                  for (i in 1:length(options))
                    if (as.integer(tclvalue(paste("checkoption", i, sep = ""))) > 0)
                        uChoice <<- c(uChoice, i)
                }
            }
        }
      else
        {
          uAnswer <- tkentry(ubox)
          tkpack(uAnswer)
          doAnswer <- function()
            {
              uChoice <<- tclvalue(tkget(uAnswer))
            }
        }

      uOK <- tkbutton(ubox, text = "OK", command = end)
      tkpack(uOK)

      uChoice <- character()
      
      tkgrab.set(ubox)
      tkwait.window(ubox)

      ## request user choice and convert to the required format
      choice <- do.call("convert",c(list(X = uChoice), cnvArgs))

      ## did the conversion fail?
      if (any(is.na(choice))) {
        tkmessageBox(title = "Wrong input",
                     message = do.call("error",
                       c(list(X = uChoice),
                         errArgs)),
                     icon = "warning",
                     type = "ok")
        next
      }
      ## Check the user choice. If the function returns TRUE,
      ## the user input is ok and can be returned
      if (do.call("check",c(list(X = choice), chkArgs))) break
      
      tkmessageBox(title = "Wrong input",
                   message = do.call("error",
                     c(list(X = choice),
                       errArgs)),
                   icon = "warning",
                   type = "ok")
    } 
    choice
  }

## A menu function for selecting an option
selectOption <- function(options, message)
  {
    choice <- selectCore(message,
                         function(X) {as.integer(X)},
                         function(X, count) {X > 0 & X <= count},
                         options = options,
                         optflag = TRUE,
                         count = length(options)
                         )
    options[choice]
  }

## A menu function for selecting several options
selectOptions <- function(options, message, all)
  {
    Z <- c(options, all)
    choice <- selectCore(message,
                         function(X) {as.integer(X)},
                         function(X, count) {length(X) > 0 & length(X) <= count},
                         options = Z,
                         optflag = FALSE,
                         count = length(Z)
                         )
    Y <- Z[choice]
    if(any(Y==all)) return(options)
    intersect(Y, Z)
  }

## A menu function for selecting a child class
selectChild <- function(X, message)
  {
    children <- getChildrenClasses(X)
    ## in case the class does not exist,
    ## the calling function should try to
    ## generate the generic class
    if (length(children) == 0) return(X)
    shortNames <- gsub(X, "", children)
    paste(X, selectOption(shortNames, message), sep = "")
  }


## A menu function for selecting a number
selectNumber <- function(message, min, max)
  {
    selectCore(paste(message, " [", min, ", ", max, "]\n\n", sep = ""),
               function(X) as.integer(X),
               function(X, min, max) X >= min & X <= max,
               min = min,
               max = max
               )
  }

## A menu function for selecting a number
selectDouble <- function(message, min, max)
  {
    selectCore(paste(message, " [", min, ", ", max, "]\n\n", sep = ""),
               function(X, ...) as.double(X),
               function(X, ...) X >= min & X <= max,
               min = min,
               max = max
               )
  }

## RFE: This could be an object method for the different
##      goCluster objects [each class specifying protected
##      slots in the "protect"-slot]

## A check function. It is being used to prevent overwriting
## already calculated results. Not very useful yet.
checkData <- function(variable, action, call, warn=FALSE)
{
  if (length(variable)!=0)
    {
      message <- paste("The object has already been ",
                       action,
                       "! Ignoring \",",
                       call,
                       ",\" call!",
                       sep = "")
      if (warn)
        {
          warning(message)
          return(TRUE)
        }
      else
        {
          stop(message)
        }

    }
  else
    {
      return(FALSE)
    }
}

printData <- function(title, type, details = NULL)
{

  cat("  ",title ,":\n", sep ="")
  cat("  ",rep("-",nchar(title) + 1), "\n\n", sep ="")
  cat("    Type:", type, "\n\n")
  if (!is.null(details))
    {
      cat("    Details:\n\n")
      for (line in details)
        {
          cat("      ->", line,"\n")
        }
      cat("\n")
    }
}

getGeneColors <- function()
{
  ## Color generation
  meanCol   <- function(X, Y, N) floor(matrix(X, 3, N) + matrix(seq(0, 1, by = 1 / (N - 1)), 3, N, byrow = TRUE) * (matrix(Y, 3, N) - matrix(X, 3, N)))
  expandCol <- function(X, N) {res <- NULL; for (i in 1:(dim(X)[2] - 1)) res <- cbind(res, meanCol(X[, i], X[, i + 1], N)[, 1:(N - 1)]); res}
  genecolors <- c("#313695", "#4575B4", "#74ADD1", "#ABD9E9", "#E0F3F8", "#FFFFBF", "#FEE090", "#FDAE61", "#F46D43", "#D73027", "#A50026")
  genecolors <- expandCol(col2rgb(genecolors),5)
  genecolors
}

## A convenience function
goClusterTk <- function()
{
  options(goCluster.useTk = 1)
  a <- new("goCluster")
  a <- config(a)
  a <- execute(a)
  a
}

##################################################################################
##
## GOCLUSTER GENERAL CLASS DEFINITIONS
##
##################################################################################

## Generic module definition (needed for the generic functions)
setClass("clusterModule",
         representation(## Interactive configuration
                        config  = "list",
                        ## Non-interactive config checks
                        setup   = "list",
                        ## Main function of the class
                        execute = "function",
                        ## A list of protected slots
                        reset   = "list",
                        ## A list of children
                        children = "list"
                        )
         )

## This serves as a container for the annotation of the dataset.
## The clusterAnnotation object is a "child" of the clusterData
## object. It should be capable of automatically retrieving the
## annotation data for the parent object.

setClass("clusterAnnotation",
         representation(
                        ## The type of annotation
                        type    = "character",
                        ## The annotations as list. Each list should
                        ## contain a two-column table. The first column
                        ## specifies the unique id, the second specifies
                        ## the associated annotation
                        annoset = "list"
                        ),
         contains = "clusterModule"
         )

setMethod("print", "clusterAnnotation", function(x, ...)
          {
            printData("Annotation","not defined")
          }
          )

## The expression dataset. Each row of the dataset (gene) needs
## a reference (unique ID) to an annotation dataset, so that
## all relevent annotation can be automatically retrieved by the
## clusterAnnotation object that is stored in the "anno" slot.

setClass("clusterData",
         representation(
                        ## An arbitrary name for the dataset
                        name    = "character",
                        ## The actual data
                        dataset = "exprSet",
                        ## Unique IDs with the same length as the
                        ## rows of the dataset
                        uniqueid = "character",
                        ## The annotation of the dataset
                        anno = "clusterAnnotation"
                        ),
         prototype(config = list(
                     name =
                     function (object)
                     {
                       selectCore("Please select a name or title for the dataset.")
                     },
                     dataset =
                     function (object)
                     {
                       get(selectCore("Please specify the variable name of the dataset to be analyzed.",
                                      check =
                                      function(X)
                                      {
                                        if (exists(X) && is(get(X), "exprSet"))
                                          return(TRUE)
                                        return(FALSE)
                                      },
                                      error =
                                      function(X)
                                      {
                                        return(paste("The specified variable (", X,
                                        ") does not exist or is no expression dataset.",
                                                     sep =""))
                                      }
                                      )
                           )
                     },
                     uniqueid =
                     function (object)
                     {
                       a <- selectCore("Please specify the variable name of the unique ID for the rows\nof the dataset. Alternatively you can specify \"rownames\" and\nthe current rownames of the dataset will be used as unique ID.",
                                       check =
                                       function(X)
                                       {
                                         if (X == "rownames") return(TRUE)
                                         if (exists(X) &&
                                             length(get(X)) == dim(exprs(object@dataset))[1])
                                           return(TRUE)
                                         return(FALSE)
                                       },
                                       error =
                                       function(X)
                                       {
                                         return(paste("The specified variable (", X,
                                                      ") does not exist or does not",
                                                      "match the size of your dataset.",
                                                      sep =""))
                                       },
                                       dataset = object@dataset
                                       )

                       if (a == "rownames")
                         {
                           object@uniqueid <- rownames(exprs(object@dataset))
                         }
                       else
                         {
                           ## FIXME: check length
                           object@uniqueid <- get(a)
                         }
                     }               
                     ),
                   setup = list(
                     name = 
                     function(X, object)
                     {
                       if (!is.character(X) || !length(X) > 0)
                           stop("The name of the dataset should be a short string!")
                     },
                     dataset = 
                     function(X, object)
                     {
                       if (!is(X, "exprSet"))
                         stop(paste("You need to provide an object of the class",
                                    "exprSet (Biobase) instead of an object of class\"",
                                    class(X), "\"!", sep = "")
                              )
                     },
                     uniqueid = 
                     function(X, object)
                     {
                       if (length(X) != dim(exprs(object@dataset))[1])
                           stop(paste("Length of the given unique id (",
                                      length(X),
                                      ") does not match the number",
                                      "of rows in the dataset (",
                                      dim(exprs(object@dataset))[1],
                                      ")!", sep = "")
                                )
                     }
                     ),
                   children = list(
                     list(slot        = "anno",
                          description = "annotation data",
                          execute     = TRUE
                          )
                     )
                   ),
         contains = "clusterModule"
         )

setMethod("print", "clusterData", function(x, ...)
          {
            if (is.null(x@dataset))
              {
                type    <- "not defined"
                details <- NULL
              }
            else
              {
                type    <- "Bioconductor ExprSet object"
                details <- capture.output(print(x@dataset))
              }
            
            printData("Dataset", type, details)
            print(x@anno)
          }
          )

## This object partitions the dataset into (possibly overlapping)
## groups. The result is stored in the clusterset slot and can be
## a list of just one group, a list of many groups or a list of lists
## with groups

setClass("clusterAlgorithm",
         representation(
                        ## The gene <-> cluster association as a list of groups.
                        ## The list may have several levels to specify
                        ## a tree of groups. This can be useful to represent
                        ## the result of a hierarchical clustering. Each level can
                        ## either be a list of list or a list of groups but not both.
                        ## The terminal
                        ## nodes of the list elements need to be a list of unique
                        ## ids specifying the genes of the group. 
                        clusterset  = "list"
                        ),
         prototype(reset = list(clusterset = list())
                   ),
         contains = "clusterModule"
         )

setMethod("print", "clusterAlgorithm", function(x, ...)
          {
            printData("Algorithm","not defined")
          }
          )

## This object calculates the statistical values for genes or annotation terms

setClass("clusterStatistic",
         representation(
                        ## This is a list ordered in the same way as the
                        ## list resulting from the clusterAlgorithm. But
                        ## here the terminal nodes hold the interesting
                        ## elements that were found in each group.
                        ## These elements could be genes or annotation terms
                        ## depending on the type of algorithm that is employed
                        ## here. 
                        statset  = "list"
                        ),
         prototype(reset = list(statset = list())
                   ),
         contains = "clusterModule"
         )

setMethod("print", "clusterStatistic", function(x, ...)
          {
            printData("Statistic","not defined")
          }
          )

## This object is meant to be used for the selection of interesting elements
## from the list that is being returned from the clusterStatistic object.
## This should include any necessary corrections for multiple testing.

setClass("clusterSignif",
         representation(
                        ## The statistical values. This may only
                        ## be a copy of the clusterStatistic result
                        ## but can also be a corrected version of
                        ## these p-values
                        statset    = "list",
                        ## A list that contains terms identified. The
                        ## structure is the same as for the clusterAlgorithm
                        ## result. Each terminal node only holds the elements
                        ## of interest.
                        selection  = "list",
                        ## This slot holds the object that will perform
                        ## the statistic. It is a subobject so that
                        ## it can be used for the analysis of randomized
                        ## datasets
                        stat       = "clusterStatistic"
                        ),
         prototype(reset = list(statset = list()),
                   children = list(
                     list(slot        = "stat",
                          description = "statistical analysis algorithm",
                          execute     = FALSE
                          )
                     )
                   ),
         contains = "clusterModule"
         )

setMethod("print", "clusterSignif", function(x, ...)
          {
            printData("Significance","not defined")
            print(x@stat)
          }
          )

## This provides functionality to visualize the clustered data

setClass("clusterVisual",
         representation(
                        ),
         contains = "clusterModule"
         )

setMethod("print", "clusterVisual", function(x, ...)
          {
            printData("Visualization","not defined")
          }
          )

## The main object joining data, analysis algorithm and visualization

setClass("goCluster",
         representation(
                        ## The data to be clustered
                        data = "clusterData",
                        ## The cluster algorithm
                        algo = "clusterAlgorithm",
                        ## The cluster significance
                        sign = "clusterSignif",
                        ## The output function
                        visu = "clusterVisual"
                        ),
         prototype(children = list(
                     list(slot        = "data",
                          description = "microarray expression data",
                          execute     = TRUE
                          ),
                     list(slot        = "algo",
                          description = "gene selection or clustering algorithm",
                          execute     = TRUE
                          ),
                     list(slot        = "sign",
                          description = "significance analysis",
                          execute     = TRUE
                          ),
                     list(slot        = "visu",
                          description = "visualization",
                          execute     = TRUE
                          )
                     )
                   ),
         contains = "clusterModule"
         )

setMethod("print", "goCluster", function(x, ...)
          {
            cat("\ngoCluster-Object:\n")
            cat("-----------------\n\n")
            print(x@data)
            cat("\n")
            print(x@algo)
            cat("\n")
            print(x@sign)
            cat("\n")
            print(x@visu)
          }
          )

##################################################################################
##
## GOCLUSTER CLASS METHOD DEFINITIONS 
##
##################################################################################


## This function is used for the interactive setup of a class
if (!isGeneric("config"))
  setGeneric("config", function(object) standardGeneric("config"))

## The corresponding non-interactive function
if (!isGeneric("setup<-"))
  setGeneric("setup<-", function(object, value) standardGeneric("setup<-"))

## This method should always return a list that can be used for the
## non-interactive setup with the setup<-method.
if (!isGeneric("setup"))
  setGeneric("setup", function(object) standardGeneric("setup"))

## This function erases all calculated datasets on an object. Setup
## and execution of the object can be run again after issuing this command.
if (!isGeneric("reset"))
  setGeneric("reset", function(object) standardGeneric("reset"))

## Execute the function of an object.
if (!isGeneric("execute"))
  setGeneric("execute", function(object, parent) standardGeneric("execute"))

## Display an object.
if (!isGeneric("display"))
  setGeneric("display", function(object, ...) standardGeneric("display"))

## Setup and run a goCluster object 
if (!isGeneric("execute<-"))
  setGeneric("execute<-", function(object, value) standardGeneric("execute<-"))

checkProtection <- function(object, method)
{
  ## check protected slots 
  resetSlots <- slot(object, "reset")
  count <- length(resetSlots)
  if (count > 0)
    {
      for (i in 1:count)
        {
          if (length(slot(object, names(resetSlots)[i]))!=length(resetSlots[[i]]))
            stop(paste("\nThe object of type ", class(object), " has already been",
                       "\nexecuted. Call a reset before running ", method, " again.",sep=""))
        }
    }
}


## Configuring a clusterModule object
setMethod("config", "clusterModule", function(object)
          {
            checkProtection(object, "config")
            ## Any data slots that should be configured?
            configSlots <- slot(object, "config")
            count <- length(configSlots)
            if (count > 0)
              {
                for (i in names(configSlots))
                  {
                    slot(object, i) <- configSlots[[i]](object)
                  }
              }
            ## Any children that also need the config call?
            children <- slot(object, "children")
            if (length(children) > 0)
              {
                for (i in children)
                  {
                    checkProtection(slot(object, i$slot), "config")
                    slot(object, i$slot) <- new(selectChild(class(slot(object, i$slot)),
                                                            paste("Please select one of the\n",
                                                                  "following classes for the\n",
                                                                  i$description, ":\n", sep="")))
                    slot(object, i$slot) <- config(slot(object, i$slot))
                  }
              }
            return(object)
          }
          )

## Return module setup
setMethod("setup", "clusterModule", function(object)
          {
            returnList <- list()
            ## What are the parameter slots?
            configSlots <- slot(object, "config")
            count <- length(configSlots)
            if (count > 0)
              {
                for (i in names(configSlots))
                  {
                    returnList[[i]] <- slot(object, i)
                  }
              }
            ## Any children that also need to be returned?
            children <- slot(object, "children")
            count <- length(children)
            if (count > 0)
              {
                for (i in children)
                  {
                    returnList[[i$slot]] <- setup(slot(object, i$slot))
                    returnList[[paste("class", i$slot, sep="")]] <- class(slot(object, i$slot))
                  }
              }
            return(returnList)
          }
          )

## Non-interactive module setup
setMethod("setup<-", "clusterModule", function(object, value)
          {
            checkProtection(object, "setup<-")
            ## What are the parameter slots?
            setupSlots <- slot(object, "setup")
            count <- length(setupSlots)
            if (count > 0)
              {
                for (i in names(setupSlots))
                  {
                    ## Call the check function to error out in
                    ## case the value is not ok
                    setupSlots[[i]](value[[i]], object)
                    ## Value ok, assign to slot
                    slot(object, i) <- value[[i]]
                  }
              }
            ## Setup the children
            children <- slot(object, "children")
            count <- length(children)
            if (count > 0)
              {
                for (i in children)
                  {
                    checkProtection(slot(object, i$slot), "setup<-")
                    className <- paste("class", i$slot, sep="")
                    if (extends(class(value[[i$slot]]),class(slot(object, i$slot))))
                      {
                        childObject <- value[[i$slot]]
                      }
                    else 
                      {
                        childObject <- try(new(value[[className]]))
                        if (inherits(childObject, "try-error"))
                          stop(paste("Could not create an object of class \"",
                                     value[[className]], "\"! Setup<- failed on",
                                     " object of class \"", class(object), "\"",
                                     " when trying to initialize child \"",
                                     i$slot , "\"", sep=""))
                        setup(childObject) <- value[[i$slot]]
                      }
                    slot(object, i$slot) <- childObject
                  }
              }
            return(object)
          }
          )

## Execute the module
setMethod("execute", "clusterModule", function(object, parent)
          {
            ## FIXME: a more generic solution
            if (is(object, "goCluster")) parent <- object

            ## All objects (beside the goCluster-object) need a parent
            if (missing(parent) || is.null(parent))
              stop(paste("\nThe object of type ", class(object), "needs to have",
                         "\n a defined parent before it can be used!", sep = ""))
                
            checkProtection(object, "execute")
            
            returnList <- list()
            ## Has an execution function been set?
            efun <- slot(object, "execute")
            if(!is.null(unlist(as.list(efun)))) object <- efun(object, parent)

            ## Any children that also need to be executed?
            children <- slot(object, "children")
            count <- length(children)
            if (count > 0)
              {
                for (i in children)
                  {
                    ## FIXME: see above. The top level needs to be updated here.
                    if (is(object, "goCluster")) parent <- object
                    if (i$execute)
                      slot(object, i$slot) <- execute(slot(object, i$slot), parent)
                  }
              }
            return(object)
          }
          )

## Setup and execute
setMethod("execute<-", "clusterModule", function(object, value)
          {
            setup(object) <- value
            execute(object, NULL)
          }
          )

## Resetting a clusterModule object
setMethod("reset", "clusterModule", function(object)
          {
            ## Any data slot that should be cleared?
            resetSlots <- slot(object, "reset")
            count <- length(resetSlots)
            if (count > 0)
              {
                for (i in 1:count)
                  {
                    slot(object, names(resetSlots)[i]) <- resetSlots[[i]]
                  }
              }
            ## Any children that also need the reset call?
            children <- slot(object, "children")
            count <- length(children)
            if (count > 0)
              {
                for (i in children)
                  {
                    slot(object, i$slot) <- reset(slot(object, i$slot))
                  }
              }
            return(object)
          }
          )

## visualize a goCluster object
setMethod("display", "goCluster", function(object, ...)
          {
            display(object@visu, object, ...)
          }
          )

